#include <kernel.h>
#include <mm/page.h>
#include <arch/mm.h>
#include <arch/idt.h>

#define _MM_PREFIX "(mm/paging): "

/* Various kernel symbols */
extern u_int32_t _text, _etext, _end, i386_endbase;
extern u_int32_t __init_start, __init_end;

p_pagedir	pd;
p_pagetbl	pt;
p_page		frames;
u_int32_t	mem_end, phys_pt;

u_int32_t get_free_pages(u_int32_t mask, int order)
{
	u_int32_t t,tbl,e;
	u_int32_t flags,m;

	if ( order ) {
		kprint(ARCH_NAME " vm unable to handle %i order page allocations\n", order);
		return 0;
	}
	
	/* Set the bits we are interested in */
	m=(1<<PTE_PRESENT) | (1<<PTE_RW) | (1<<PTE_LOCKED) | (1<<PTE_DMA);

	/* Setup the flags we need */
	flags=mask;
	MASK_SET(flags, PTE_PRESENT);
	MASK_SET(flags, PTE_RW);
	MASK_UNSET(flags, PTE_LOCKED);
	flags=flags&m;

	for(t=0; t<phys_pt; t++){
		tbl=pd->entry[t].value;
		if ( tbl & (1<<PDE_PRESENT) ) {
			tbl=__dir(tbl);
		}else{
			continue;
		}

		for(e=0; e<1024; e++) {
			if ( (pt[tbl].entry[e].value & m) == flags ) {
				/* Lock in this page */
				MASK_SET(pt[tbl].entry[e].value, PTE_LOCKED);

				/* Return virtual address */
				return __mkvirt(tbl, e, 0);
			}
		}
	}
	
	return 0;
}

/* Locks physical address */
void lock_page(u_int32_t addr)
{
	u_int32_t tbl=__dir(pd->entry[dir(addr)].value);
	if ( ofs(addr) ) {kprint(_MM_PREFIX "WARNING: lock_page() at offset\n"); };
	MASK_SET(pt[tbl].entry[tbl(addr)].value, PTE_LOCKED);
}

/* Unlocks physical address */
void unlock_page(u_int32_t addr)
{
	u_int32_t tbl=__dir(pd->entry[dir(addr)].value);
	if ( ofs(addr) ) {kprint(_MM_PREFIX "WARNING: unlock_page() at offset\n"); };
	MASK_UNSET(pt[tbl].entry[tbl(addr)].value, PTE_LOCKED);
}

void lock_region(char *name, u_int32_t begin, u_int32_t end)
{
	u_int32_t p;
	int kb=(end-begin)/1024;

	/* Round up and down to pages */
	if ( begin & (0xFFFFFFFF>>20) ) {
		begin-=4096;
		begin&=0xFFFFFFFF<<12;
	}
	if ( end & (0xFFFFFFFF>>20) ) {
		end+=4096;
		end&=0xFFFFFFFF<<12;
	}
	
	for(p=begin; p<end; p+=PAGE_SIZE) {
		lock_page(p);
	}
	/*
	kprint(_MM_PREFIX "%s ", name);
	koutput.xpos=28; kprint("0x%x", begin);
	koutput.xpos=39; kprint("-> 0x%x", end);
	koutput.xpos=52; kprint("(%u %s)\n",
		kb>1024 ? kb/1024 : kb,
		kb>1024 ? "MB" : "KB");
	*/
}

void __init paging_init(void)
{
	u_int32_t pds;
	u_int32_t i,j;
	u_int32_t pfa;
	u_int32_t flags;
	
	/* Figure out memory size etc.. */
	mem_end=boot_info.mem_lo+boot_info.mem_up;
	mem_end *= 1024; /* Get bytes */
	mem_end &= 0xFFFFFFFF<<12; /* Mask off pages */
	phys_pt=(mem_end/(PAGE_SIZE*MAX_PAGE_TBL_ENTRIES))+1;
	mem_end=phys_pt*(PAGE_SIZE*MAX_PAGE_TBL_ENTRIES);
	boot_info.mem_size=mem_end;
	
	/* Locate the page frame array */
	pfa=(u_int32_t)&_end;
	if ( pfa % PAGE_SIZE ) pfa+=PAGE_SIZE;
	pfa&=0xFFFFFFFF<<12;
	frames=(p_page)pfa;
	
	/* Now locate the PDE */
	pds=(sizeof(page_t)*(phys_pt*MAX_PAGE_TBL_ENTRIES));
	if ( pds % PAGE_SIZE ) pds+=PAGE_SIZE;
	pds&=0xFFFFFFFF<<12;
	pds+=pfa;

	/* Store the results */
	pd=(p_pagedir)pds;
	pt=(p_pagetbl)(pds+PAGE_SIZE);
	
	/* Fill in the page frame descriptors */
	for(i=0; i<phys_pt*MAX_PAGE_TBL_ENTRIES; i++) {
		frames[i].count=0;
		frames[i].phys=i*PAGE_SIZE;
	}
	
	/* Blank the page directory */
	for(i=0; i<MAX_PAGE_DIR_ENTRIES; i++) {
		pd->entry[i].value=0;
	}

	/* Mark present all physical pages */
#ifdef CONFIG_SMALL_MEM
	for(i=0; i<phys_pt; i++) {
#else
	for(i=0; i<MAX_PAGE_DIR_ENTRIES; i++) {
#endif
		int jinit=0;
		u_int32_t addr;
		
		pd->entry[i].value = ((u_int32_t)&pt[i]) | (1<<PDE_PRESENT) | (1<<PDE_RW);
		
		/* Make an exception to catch NULL dereference */
		if ( i==0 ) jinit=1;
		
		for(j=jinit; j<MAX_PAGE_TBL_ENTRIES; j++) {
			flags = (1<<PTE_PRESENT) | (1<<PTE_RW);
			addr = ((j+(PAGE_SIZE*i))<<12);

			if ( addr < 0x100000 ) {
				MASK_SET(flags, PTE_DMA);
			}
			
			pt[i].entry[j].value = addr|flags;
		}
	}

}

void __init paging_enable(void)
{
	u_int32_t x,y;

	/* set page directory register to our PD */
	asm volatile ("movl %%edx, %%cr3":: "d" ((u_int32_t)pd));

	/* Enable PG (paging) bit in cr0 */
	asm volatile (
	"movl	%cr0, %eax		\n"
	"orl	$0x80000000, %eax	\n"
	"movl	%eax, %cr0		\n");

	/* Setup page fault handler */
	idt_trap(pagefault, 0x0e);
	
	/* Reserve memory regions */
	x=(u_int32_t)pd;
#ifdef CONFIG_SMALL_MEM
	y=(u_int32_t)pd+PAGE_SIZE+(phys_pt*PAGE_SIZE);
#else
	y=(u_int32_t)pd+PAGE_SIZE+(MAX_PAGE_DIR_ENTRIES*PAGE_SIZE);
#endif
	lock_region("BIOS Data", 0, 4096); /* Catches NULL ptr anyway */
	lock_region("ISA VGA", 0xa0000, 0x100000);
	lock_region("Kernel", (u_int32_t)&_text, (u_int32_t)&_end);
	lock_region("Paging data", x, y);

	/* Print some stats */
	kprint(_MM_PREFIX "Physical mem boundary: 0x%x\n",mem_end);
	kprint(_MM_PREFIX "Paging dirs/tables: %uKB (%u pages present)\n",
		(y-x)/1024, phys_pt*MAX_PAGE_TBL_ENTRIES);
	kprint(_MM_PREFIX "%uKB of init data to discard\n", 
		((u_int32_t)&__init_end-(u_int32_t)&__init_start)/1024);
}
