#include <kernel.h>
#include <arch/setup.h>
#include <arch/descriptor.h>
#include <arch/8259a.h>
#include <arch/keyboard.h>
#include <arch/smp.h>
#include <arch/timer.h>
#include <arch/dma.h>
#include <arch/fdc.h>
#include <arch/pci.h>
#include <arch/mm.h>
#include <lib/time.h>

EXPORT_SYMBOL(cli);
EXPORT_SYMBOL(sti);
EXPORT_SYMBOL(reboot);
EXPORT_SYMBOL(enter_to_exit);

char __initdata *monitor_str[]= {
	"Not Present",
	"Colour VGA Monitor",
	"Monochrome VGA (High contrast output enabled)",
	"Unknown (Non standard VGA adapter?)"
};

driver_t __initdata pic[]={
	{"8259A PIC", pic_init, pic_probe, pic_setup},
	/* TODO: APIC */
};

driver_t __initdata pc_driver[]={
	{"BIOS VGA access", NULL, NULL, NULL}, 
	{"PC/AT Keyboard", keyboard_init, keyboard_probe, keyboard_setup},
	{"DMA Controller", dma_init, dma_probe, dma_setup},
	{"PCI busses", pci_init, pci_probe, pci_setup},
	{"AT Floppy", fdc_init, fdc_probe, fdc_setup}, 
};

/* Global descriptor table */
dt_entry_t GDT[]={
	{dummy:0},

	/* Kernel space */
	stnd_desc(0, 0xFFFFF, (D_CODE | D_READ  | D_BIG | D_BIG_LIM)),
	stnd_desc(0, 0xFFFFF, (D_DATA | D_WRITE | D_BIG | D_BIG_LIM)),

	/* User space */
	stnd_desc(0, 0xFFFFF, (D_CODE | D_READ  | D_BIG | D_BIG_LIM | D_DPL3)),
	stnd_desc(0, 0xFFFFF, (D_DATA | D_WRITE | D_BIG | D_BIG_LIM | D_DPL3)),
};
struct {
	unsigned short limit __attribute__((packed));
	union dt_entry *gdt __attribute__((packed));
}loadgdt={ (3*sizeof(dt_entry_t)-1), GDT};

void __init print_cpu_info(u_int32_t cpu)
{
	u_int32_t i=0;
	
	kprint("     (cpu%u): %u.%u MHz. ", 
		cpu,
		((delayinfo.lpms*10)/5000),
		((delayinfo.lpms*10)/50)%100);

	while(features[i++].mask){
		if ( features[i-1].mask & smpinfo.cpu[cpu].features ) {
			kprint("(%s) ", features[i-1].name);
		}
	}
	
	kprint("\n");
}

/* Initialises all architecture specific drivers etc..
* then passes control to the platform-independant kernel_main() */
void __init setup (void)
{
	int i=0;
	
	/* Cute banner and boot info */
	kprint("ScaraOS for " ARCH_NAME_L  " Architecture (" ARCH_NAME  ")\n");
	paging_init();
	
	/* Stuff in the boot_info struct 
	* (that the bootloader setup for us)*/
	kprint("     (boot): %s\n", boot_str[boot_info.boot_method]);
	kprint("  (monitor): %s\n", monitor_str[boot_info.monitor]);
	kprint("  (cmdline): %s\n", boot_info.cmd_line);
	kprint("   (memory): %u.%uMB: (%uKB kernel)\n",
		(boot_info.mem_size/(1024*1024)),
		((boot_info.mem_size/(1024*1024))/100)%100,
		(boot_info.kernel_end-boot_info.kernel_addr)/1024);
	
	asm volatile ("lgdtl (loadgdt)\n"); /* Load GDT   */

	/* We need to probe for SMP early on so we
	* know whether to use PIC or APIC for IRQs */
	boot_info.smp=probe_smp();

	/* Need the PIC before paging, cos we need the IDT */
	pic[0].init();
	pic[0].probe();
	pic[0].setup();
	
	/* Start paging */
	paging_enable();
	
	/* Calibrate delay loop */
	pit_init();

	/* TODO: CPU detection */
	print_cpu_info(smpinfo.boot_cpu);

	/* Initialise all low level IA32 drivers (PC/AT shit basically) */
	for(i=0; i<(sizeof(pc_driver)/sizeof(driver_t)); i++) {
		u_int32_t p=0,s=0;
		if ( pc_driver[i].init ) pc_driver[i].init();
		if ( pc_driver[i].probe ) p=pc_driver[i].probe();
		if ( pc_driver[i].setup ) s=pc_driver[i].setup();

		if ( pc_driver[i].probe && pc_driver[i].setup ) {
			//kprint("   (driver): %s: %u/%u Devices Configured\n",
			//	pc_driver[i].name, s, p);
		}else{
			//kprint("   (driver): %s: Device Configured\n",
			//	pc_driver[i].name, s, p);
		}
	}

	/* That wasn't too painful, we have now initialise pretty much
	* all of the IA32 dependent drivers etc, now to get a couple of
	* niggles out of the way so we can go to the kernel proper  */

	/* Start the jiffie timer */
	pit_start_timer1();

	/* Main architecture independent kernel */
	kernel_main();
	return;
}

/* Wrappers for cli and sti instructions */
void cli(void){asm("cli\n");};
void sti(void){asm("sti\n");};
void reboot(void)
{
	/* Reset the machine using the KB contoller */
	while(inb(0x64) & 0x02);
	outb(0x64, 0xfe);

	/* If it doesn't work, at least make sure the CPU
	* stays cool ;) */
	cli();
	asm(
		"j1:	hlt		\n"
		"jmp	j1		\n"
	);
}

/* Code for enter_to_exit() */
void exit_handler(void)
{
	u_int8_t k=inb_p(0x60);
	
	if ( k == 0x1c ) {
		reboot();
	}
	irq_eoi(1);

	return;
}
void exit_kb_isr(void);
asm(
	".globl exit_kb_isr		\n"
	"exit_kb_isr:			\n"
	"	pusha			\n"
	"	call exit_handler	\n"
	"	popa			\n"
	"	iret			\n"
);
void enter_to_exit(void)
{
	kprint("\nPress <ENTER> to reboot: ");
	idt_interrupt(exit_kb_isr, VECTOR_M+1);
	irq_on(1);
	sti();
	while(1){asm("hlt");}; /* Keep CPU cool */
}
